#include "ThreadPool.hpp"

ThreadPool::ThreadPool(size_t nThreads) : threads{}, tasks{}, mutex{}, activity{}, inactive{}, end{false}, actives{0}
{
  auto consummer = [this] () {
		     // First lock the mutex to protect from race conditions
		     std::unique_lock<std::mutex> lock(mutex);

		     while(true) {
		       // If a task is waiting in the queue
		       if(! tasks.empty()) {
			 // Retrieve the task (mutex is locked so it is safe)
			 function_t task = tasks.front();
			 tasks.pop_front();

			 // The thread is going to be active
			 ++actives;
			 // Release the lock before running the task
			 lock.unlock();
			 // Run the task
			 task();
			 // Wait to get exclusive access to the queue again
			 lock.lock();
			 // The thread is not active anymore
			 --actives;

			 // Signal other threads if there is no more tasks to run
			 if(tasks.empty())
			   inactive.notify_all();

		       } else if(end) {
			 // If there is no more task to run and if the pool destruction is pending
			 // Terminate the thread
			 break;
		       } else {
			 // If there is no more task to run
			 // Wait for a new task to be posted
			 activity.wait(lock);
		       }
		     }
		   };
  while(nThreads-- != 0)
    // Launch a new thread to run the consummer lambda function
    threads.emplace_back(consummer);
  }

ThreadPool::~ThreadPool() {
  // Destructor of the pool
  // Tells to thread to stop as soon as possible
  end = true;

  // Wake up all waiting threads
  activity.notify_all();

  // Finally wait for the termination of all threads
  for(auto& thread : threads) thread.join();
}

void ThreadPool::join() {
  // Wait that all threads have finished

  // Lock the mutex
  std::unique_lock<std::mutex> lock(mutex);

  // While loop to protect from spurious wakeups
  while(! tasks.empty() || actives > 0)
    inactive.wait(lock);
}	
