#pragma once

#include <thread>
#include <mutex>
#include <condition_variable>
#include <functional>
#include <vector>
#include <deque>

class ThreadPool {
  using function_t = std::function<void()>;

  std::vector<std::thread> threads;	// Pool of threads
  std::deque<function_t> tasks;		// Queue containing pending tasks
  std::mutex mutex;			// Mutex to protect the task queue
  std::condition_variable activity;	// Condition variable to signal a new message has been posted
  std::condition_variable inactive;	// Condition variable to signal no more tasks are currently processed
  bool end;				// Flag to indicate to threads that they should end (the pool is being destroyed)
  unsigned int actives;			// Number of active threads
public:
  struct destroyed_pool {};
  
  // Constructor of a pool with a given number of threads
  ThreadPool(size_t nThreads);

  // Destructor to remove ressources
  ~ThreadPool();

  // Method template to post any callable with its arguments
  template<typename Func, typename... Args>
  ThreadPool& operator() (const Func& func, Args... args) {
    // Post a new task
    {
      // This is a critical section protected by a mutex
      std::lock_guard<std::mutex> lock(mutex);

      // This case should normally not occur
      if(end) throw destroyed_pool{};

      // Post the new task into the queue
      tasks.emplace_back(func, args...);
    }
    // Wake up a single thread (if any)
    activity.notify_one();

    return *this;
  }

  // Method to join all threads of the pool
  void join();
};
